from functools import partial

import jax.numpy as jnp
import jax
from ludax.config import ACTION_DTYPE, REWARD_DTYPE

@jax.jit
def zero_heuristic(state_b, *_):
    """
    A zero heuristic that returns 0.0 for all states.
    """

    return jnp.zeros(state_b.legal_action_mask.shape[0], dtype=REWARD_DTYPE)


def construct_playout_heuristic(step_b, num_playouts=20):
    """
    A heuristic that performs random playouts from the current state and returns the average outcome from the
    perspective of the current player.
    """

    def cond_fn(carry):
        state, _key = carry
        # continue while at least one playout is not terminated
        return jnp.any(~state.terminated)

    def body_fn(carry):
        state, _key = carry
        _key, subkey = jax.random.split(_key)

        legal_action_mask = state.legal_action_mask  # [N, A] after repeat
        # build logits: 0 for legal moves, -inf for illegal
        logits = jnp.where(legal_action_mask.astype(bool), 0.0, -jnp.inf)  # [N, A]

        # sample one action per state (remove the last axis → shape [N])
        action = jax.random.categorical(subkey, logits=logits, axis=-1).astype(ACTION_DTYPE)

        state = step_b(state, action)
        return state, _key

    def random_playout_heuristic_multi(state_b, key):

        # Repeat each element of the batch num_playouts times along axis 0
        state_repeated = jax.tree_util.tree_map(
            lambda x: jnp.repeat(x, num_playouts, axis=0), state_b
        )

        final_state_repeated, _ = jax.lax.while_loop(cond_fn, body_fn, (state_repeated, key))

        # Average rewards over playouts, from each original state's current player perspective
        rewards = final_state_repeated.rewards                     # [N, num_players]
        current_players = state_repeated.game_state.current_player # [N]
        batch_size = state_b.legal_action_mask.shape[0]
        N = rewards.shape[0]  # batch_size * num_playouts

        rewards_from_perspective = rewards[jnp.arange(N), current_players]  # [N]
        avg_rewards = rewards_from_perspective.reshape(batch_size, num_playouts).mean(axis=1)

        return avg_rewards

    def random_playout_heuristic_single(state_b, key):
        final_state, _ = jax.lax.while_loop(cond_fn, body_fn, (state_b, key))
        return final_state.rewards[jnp.arange(final_state.rewards.shape[0]), state_b.game_state.current_player]

    if num_playouts == 1:
        return jax.jit(random_playout_heuristic_single)

    return jax.jit(random_playout_heuristic_multi)






import jax
import jax.numpy as jnp
from jax import lax

from . import BIG, SMALL, zero_heuristic
from ludax.config import ACTION_DTYPE

from .simple import random_policy

def beam_search_policy(step_b, heuristic=zero_heuristic, cache_size=1000, topk=100, iterations=10):
    # Temporary for debugging
    random_policy_f = random_policy()

    def beam_search_f(state, key):
        """Expand the most important actions at each iteration. Then run minimax on the most important states."""
        """
        Temporary beam search loop with debug output using jax.debug.print.
        Calls beam_step_f_single, prints debug info, and picks a random valid action at the end.
        """
        root_player = state.game_state.current_player

        # Start with the initial state repeated topk times
        frontier_states = jax.tree_util.tree_map(
            lambda x: jnp.repeat(jnp.expand_dims(x, axis=0), topk, axis=0),
            state
        )

        root_value = heuristic(state) # TODO figure out if it's okay to expect heuristic to support both single and batched states
        frontier_values = jnp.full((topk,), -jnp.inf, dtype=root_value.dtype)
        frontier_values = frontier_values.at[0].set(root_value)

        jax.debug.print("[Init] frontier_values = {x}", x=frontier_values)

        for i in range(iterations):
            key, subkey = jax.random.split(key)
            top_states, top_response_values, new_frontier_states, new_frontier_values = beam_step_f_single(
                root_player, frontier_states, frontier_values, subkey
            )
            jax.debug.print("[Iter {i}] frontier_values = {x}", i=i + 1, x=frontier_values)
            jax.debug.print("           top_response_values = {x}", i=i + 1, x=top_response_values)

        return random_policy_f(state, key)


    def beam_step_f_single(root_player, frontier_states, frontier_values, key):
        """
        Perform one step of the beam search, expanding the top-k states and evaluating their actions.
        Top-k is defined from the perspective of the mover sice we want to identify
        """

        # Pop the top-k frontier states and leave the rest in cache.
        # Assumes values_b is from the perspective of the mover to the state. This is not good for the minmax tree since
        # state doesn't have a reference to who was the previous player. But values_b is only used to select which
        # states to expand, as such it's only trying to prioritize exploring good actions for the previous player.
        # The top_response_values are the values are instead used for minimax.
        top_values, top_idx = lax.top_k(frontier_values, topk)
        top_states = jax.tree_util.tree_map(
            lambda x: jnp.take(x, top_idx, axis=0), frontier_states
        )

        # Expand the frontier states
        _topk, num_actions = top_states.legal_action_mask.shape

        all_actions = jnp.broadcast_to(jnp.arange(num_actions), (topk, num_actions))
        actions_flat = all_actions.reshape(-1)

        state_flat = jax.tree_util.tree_map(
            lambda x: jnp.repeat(x, num_actions, axis=0), top_states
        )

        flat_next_state = step_b(state_flat, actions_flat.astype(ACTION_DTYPE))

        # Small noise to break ties
        noise = jax.random.uniform(key, shape=flat_next_state.mover_reward.shape)

        # Signs per term
        heur_sign = jnp.where(flat_next_state.current_player == state_flat.current_player, 1.0, -1.0)

        # Combine from current player's perspective
        action_values_flat = (
                BIG * flat_next_state.mover_reward + heur_sign * heuristic(flat_next_state) + SMALL * noise
        )

        # Mask out dummy children by setting their values to the parent's value if the parent's value is <= -BIG.
        top_values_flat = jnp.repeat(top_values, num_actions)
        action_values_flat = jnp.where(top_values_flat == -jnp.inf, top_values_flat, action_values_flat)

        # Mask illegal actions as -inf
        legality_flat = state_flat.legal_action_mask[
            jnp.arange(actions_flat.shape[0]), actions_flat
        ]
        action_values_flat = jnp.where(legality_flat, action_values_flat, -jnp.inf)


        # Identify the value of the best response to each explored action.
        # This will be used to in the minmax tree, but won't determine which nodes will be expanded next.
        # top_response_values = jnp.zeros_like(top_states.legal_action_mask.shape) # temporary placeholder
        top_response_values = jnp.max(
            action_values_flat.reshape((topk, num_actions)), axis=1
        )
        # Make response values from the perspective of the root not the mover
        root_sign = jnp.where(state_flat.current_player == root_player, 1.0, -1.0)
        top_response_values = root_sign * top_response_values

        # jax.debug.print(
        #     "[beam_step] action_values_flat = {x}",
        #     x=action_values_flat,
        #     ordered=True
        # )

        # Cache states that are not in the top-k
        n = frontier_values.shape[0]
        is_top = jnp.zeros(n, dtype=bool).at[top_idx].set(True)  # mark tops
        rest_idx = jnp.nonzero(~is_top, size=n - topk, fill_value=0)[0]

        cache_values = jnp.take(frontier_values, rest_idx, axis=0)
        cache_states = jax.tree_util.tree_map(
            lambda x: jnp.take(x, rest_idx, axis=0), frontier_states
        )

        # Build combined pool to select the top-k from.
        combined_states = jax.tree_util.tree_map(
            lambda old, new: jnp.concatenate([old, new], axis=0),
            cache_states, flat_next_state
        )
        combined_values = jnp.concatenate([cache_values, action_values_flat], axis=0)

        # Select new top states and values for the cache.
        _top_val, top_indices = lax.top_k(combined_values, cache_size)

        new_frontier_values = jnp.take(combined_values, top_indices, axis=0)
        new_frontier_states = jax.tree_util.tree_map(
            lambda x: jnp.take(x, top_indices, axis=0), combined_states
        )

        # Debugging
        jax.debug.print(
            "[beam_step] legal_pairs={lp}/{tot}, kept_min={mn}, kept_max={mx}",
            lp=jnp.sum(legality_flat), tot=legality_flat.size,
            mn=jnp.min(new_frontier_values), mx=jnp.max(new_frontier_values)
        )

        return top_states, top_response_values, new_frontier_states, new_frontier_values


    return jax.jit(jax.vmap(beam_search_f, in_axes=(0, None), out_axes=0))


from typing import Any
try:
    from enum import StrEnum  # Python ≥ 3.11
except ImportError: # Python ≤ 3.10
    from enum import Enum
    class StrEnum(str, Enum):
        """Minimal stand-in for the 3.11 enum.StrEnum."""
        def _generate_next_value_(name, start, count, last_values):
            return name.lower()

import jax.numpy as jnp

from .struct import dataclass

Array = Any
PRNGKey = Any

# Configurable dtypes for quantization profiling
BOARD_DTYPE = jnp.int8     # board cells, piece/player indices, integer masks
ACTION_DTYPE = jnp.int16   # action indices, step/phase counters
REWARD_DTYPE = jnp.float32 # rewards and value estimates

TRUE = jnp.bool_(True)
FALSE = jnp.bool_(False)

INVALID = BOARD_DTYPE(-2)  # used to represent invalid cells when projecting hex grid onto rectangular array
EMPTY = BOARD_DTYPE(-1)
P1 = BOARD_DTYPE(0)
P2 = BOARD_DTYPE(1)

MAX_STEP_COUNT = 2000

@dataclass
class State():
    '''
    The current state of the environment

    Attributes:
    - game_state: a custom class that contains all the information necessary to advance the game state (e.g. board, current player, ...)
    - current_player: the identity of the current player (copied from game_state for convenience)
    - legal_action_mask: a mask indicating which actions are legal
    - winner: the player who won the game
    - rewards: the rewards for each player
    - mover_reward: the reward for the player who made the last move
    - terminated: whether the game has ended
    - truncated: whether the game has been truncated for lasting too many steps
    - global_step_count: the number of steps taken in the game
    '''
    game_state: type
    current_player: Array
    legal_action_mask: Array
    winners: Array = EMPTY * jnp.ones(2, BOARD_DTYPE)
    rewards: Array = jnp.array([0.0, 0.0], dtype=REWARD_DTYPE)
    mover_reward: Array = REWARD_DTYPE(0.0)
    terminated: Array = FALSE
    truncated: Array = FALSE
    global_step_count: Array = ACTION_DTYPE(0)

class ActionTypes(StrEnum):
    TO = 'action_to'
    FROM_DIR = 'action_from_dir'
    FROM_TO = 'action_from_to'

class Shapes(StrEnum):
    SQUARE = 'square_shape'
    RECTANGLE = 'rectangle_shape'
    HEXAGON = 'hexagon_shape'
    HEX_RECTANGLE = 'hex_rectangle_shape'

class EdgeTypes(StrEnum):
    TOP = 'top'
    BOTTOM = 'bottom'
    LEFT = 'left'
    RIGHT = 'right'

    # Hexagonal edges
    TOP_LEFT = 'top_left'
    TOP_RIGHT = 'top_right'
    BOTTOM_LEFT = 'bottom_left'
    BOTTOM_RIGHT = 'bottom_right'

class Directions(StrEnum):
    ANY = 'any'
    BACK_DIAGONAL = 'back_diagonal'
    DOWN = 'down'
    DOWN_LEFT = 'down_left'
    DOWN_RIGHT = 'down_right'
    FORWARD_DIAGONAL = 'forward_diagonal'
    DIAGONAL = 'diagonal'
    HORIZONTAL = 'horizontal'
    LEFT = 'left'
    ORTHOGONAL = 'orthogonal'
    RIGHT = 'right'
    UP = 'up'
    UP_LEFT = 'up_left'
    UP_RIGHT = 'up_right'
    VERTICAL = 'vertical'

class RelativeDirections(StrEnum):
    FORWARD = 'forward'
    BACKWARD = 'backward'
    FORWARD_LEFT = 'forward_left'
    FORWARD_RIGHT = 'forward_right'
    BACKWARD_LEFT = 'backward_left'
    BACKWARD_RIGHT = 'backward_right'

class Orientations(StrEnum):
    ANY = 'any'
    BACK_DIAGONAL = 'back_diagonal'
    FORWARD_DIAGONAL = 'forward_diagonal'
    DIAGONAL = 'diagonal'
    HORIZONTAL = 'horizontal'
    ORTHOGONAL = 'orthogonal'
    VERTICAL = 'vertical'

class PieceShapes(StrEnum):
    CIRCLE = 'circle'
    SQUARE = 'square'
    TRIANGLE = 'triangle'
    STAR = 'star'
    DIAMOND = 'diamond'

class PlayPhases(StrEnum):
    SPECIFIC_PLAYER = 'phase_specific_player'
    ALTERNATE = 'phase_alternate'
    ALTERNATURE_UNTIL = 'phase_alternate_until'

class PlayTypes(StrEnum):
    MOVE = 'play_move'
    PLACE = 'play_place'
    REMOVE = 'play_remove'

class PlayerAndMoverRefs(StrEnum):
    MOVER = 'mover'
    OPPONENT = 'opponent'
    P1 = 'P1'
    P2 = 'P2'
    BOTH = 'both'

class GameResult(StrEnum):
    WIN = 'result_win'
    LOSS = 'result_lose'
    DRAW = 'result_draw'
    BY_SCORE = 'result_by_score'

class Functions(StrEnum):
    CONNECTED = 'function_connected'
    LINE = 'function_line'
    PATTERN = 'function_pattern'

class Masks(StrEnum):
    ADJACENT = 'mask_adjacent'
    CORNER_CUSTODIAL = 'mask_corner_custodial'
    CUSTODIAL = 'mask_custodial'
    EMPTY = 'mask_empty'
    LOOP = 'mask_loop'
    OCCUPIED = 'mask_occupied'

class MoveTypes(StrEnum):
    PLACE = 'move_place'
    HOP = 'move_hop'
    SLIDE = 'move_slide'
    STEP = 'move_step'

# A bit odd, but pieces are either "any" or are defined by name elsewhere
class PieceRefs(StrEnum):
    ANY = 'any'

class PlayEffects(StrEnum):
    CAPTURE = 'effect_capture'
    EXTRA_TURN = 'effect_extra_turn'
    FLIP = 'effect_flip'
    PROMOTE = 'effect_promote'

class Predicates(StrEnum):
    CAN_HOP = 'predicate_can_hop'
    EXISTS = 'predicate_exists'

class OptionalArgs(StrEnum):
    CAPTURE = 'capture_arg'
    DIRECTION = 'direction_arg'
    DISTANCE = 'distance_arg'
    EXACT = 'exact_arg'
    EXCLUDE = 'exclude_arg'
    HOP_OVER = 'hop_over_arg'
    INCREMENT_SCORE = 'increment_score_arg'
    INDICES = 'indices_arg'
    ORIENTATION = 'orientation_arg'
    MOVER = 'mover_arg'
    MULTI_MASK = 'multi_mask_arg'
    PATTERN = 'pattern_arg'
    PIECE = 'piece_arg'
    PLAYER = 'player_arg'
    PRIORITY = 'priority_arg'
    ROTATE = 'rotate_arg'
    SAME_PIECE = 'same_piece_arg'
    SHAPE = 'shape_arg'

DEFAULT_ARGUMENTS = {
    Functions.CONNECTED: {OptionalArgs.PIECE: 'any', OptionalArgs.MOVER: 'mover', OptionalArgs.DIRECTION: 'any'},
    Functions.LINE: {OptionalArgs.ORIENTATION: 'any', OptionalArgs.EXACT: False, OptionalArgs.PLAYER: 'mover', OptionalArgs.EXCLUDE: None},
    Functions.PATTERN: {OptionalArgs.ROTATE: False, OptionalArgs.PLAYER: 'mover', OptionalArgs.EXCLUDE: None},

    Masks.ADJACENT: {OptionalArgs.DIRECTION: 'any'},
    Masks.CORNER_CUSTODIAL: {OptionalArgs.MOVER: 'mover'},
    Masks.CUSTODIAL: {OptionalArgs.MOVER: 'mover', OptionalArgs.ORIENTATION: 'any'},
    Masks.LOOP: {OptionalArgs.MOVER: 'mover'},
    Masks.OCCUPIED: {OptionalArgs.MOVER: 'mover'},

    MoveTypes.HOP: {OptionalArgs.DIRECTION: 'any', OptionalArgs.PIECE: 'any', OptionalArgs.HOP_OVER: 'both', OptionalArgs.CAPTURE: False, OptionalArgs.PRIORITY: 0},
    MoveTypes.SLIDE: {OptionalArgs.DIRECTION: 'any', OptionalArgs.DISTANCE: None, OptionalArgs.PRIORITY: 0},
    MoveTypes.STEP: {OptionalArgs.DIRECTION: 'any', OptionalArgs.PRIORITY: 0},

    PlayEffects.CAPTURE: {OptionalArgs.MOVER: 'opponent', OptionalArgs.INCREMENT_SCORE: False},
    PlayEffects.EXTRA_TURN: {OptionalArgs.SAME_PIECE: False},
    PlayEffects.FLIP: {OptionalArgs.MOVER: 'opponent'},
    PlayEffects.PROMOTE: {OptionalArgs.MOVER: 'mover'},

    Predicates.CAN_HOP: {OptionalArgs.DIRECTION: 'any', OptionalArgs.PIECE: 'any', OptionalArgs.MOVER: 'both'},
}


RENDER_CONFIG = {
    "cell_size": 50,
    "piece_radius": 35,
    "legal_radius": 15,
    "hexagon_orientation": "pointy",

    # Colors
    "light_blue": "#d2e6ff",
    "light_grey": "#c3cdd8",
    "white": "#fafafa",
    "black": "#323232",
    "dark_grey": "#a6a6a6",
    "purple": "#8a7af0",

    # Region colors (cycled through for multiple regions)
    "region_colors": [
        "#ffe4b5",  # light orange
        "#e4ffb5",  # light lime
        "#ffb5e4",  # light pink
        "#b5ffb5",  # light green
        "#e4b5ff",  # light purple
    ],

    # SVG Style information
    "svg_style": """
        <style>
        .fade-in-out {
            animation: fadeOutIn 2s ease-in-out;
        }
        @keyframes fadeOutIn {
            0% { opacity: 1; }
            50% { opacity: 0; }
            100% { opacity: 1; }
        }
        </style>
    """
}


BIG = 10**6
SMALL = 10**-3


from .value import zero_heuristic, construct_playout_heuristic
from .beam import beam_search_policy
from .negamax import negamax_policy
from .simple import one_ply_policy, random_policy
from .mctx import simple_mctx_policy, lookahead_mctx_policy
from .mcts import uct_mcts_policy


__all__ = [
    "beam_search_policy",
    "negamax_policy",
    "one_ply_policy",
    "random_policy",
    "simple_mctx_policy",
    "lookahead_mctx_policy",
    "zero_heuristic",
    "BIG",
    "SMALL"
]


'''
An extremely vanilla implementation of UCB Monte Carlo Tree Search (MCTS) in JAX / Ludax, emphasizing clarity over performance.

Largely inspired by: https://github.com/chrisgrimm/muzero/blob/main/networks/mcts.py
'''


from typing import NamedTuple, Tuple

import jax
import jax.numpy as jnp

from ludax import LudaxEnvironment
from ludax.config import State, ACTION_DTYPE, REWARD_DTYPE


REWARD_SCALE = 1
class MCTSParams(NamedTuple):
    '''
    The main data structure that stores the current state of the MCTS tree

    Params:
    - node_num (int): the current number of nodes in the tree / index of the next node to be added
    - transitions (array of shape [num_sims, num_actions]): for each node, the index of the child node for each action, or -1 if that action has not been taken yet
    - states (array of shape [num_sims, state_dim]): the game state at each node, used for expansion via the environment's step function
    - legal_actions (array of shape [num_sims, num_actions]): the legal action mask at each node
    - visits (array of shape [num_sims, num_actions]): the visit count N(s,a) for each node / action
    - rewards (array of shape [num_sims, num_actions]): the total reward W(s,a) for each node / action
    - to_play (array of shape [num_sims,]): the player to play at each node (used to flip the sign of rewards for the opponent)
    '''
    player_idx: int
    node_num: int
    transitions: jnp.ndarray
    states: jnp.ndarray
    legal_actions: jnp.ndarray
    visits: jnp.ndarray
    rewards: jnp.ndarray
    to_play: jnp.ndarray

class MCTSTraversal(NamedTuple):
    '''
    Stores the information from a single rollout from the root to a leaf node

    Params:
    - nodes (array of shape [max_depth,]): the indices of the nodes visited during the rollout
    - actions (array of shape [max_depth,]): the actions taken at each node during the rollout
    - valid (array of shape [max_depth,]): whether each step in the rollout was valid (i.e. did not reach a terminal / leaf state)
    '''
    nodes: jnp.ndarray
    actions: jnp.ndarray
    valid: jnp.ndarray

def initialize(environment: LudaxEnvironment, root_state: State, num_sims: int, max_depth: int, key) -> Tuple[MCTSParams, jnp.ndarray]:
    '''
    Initialize the MCTS tree with the root node

    Params:
    - environment (LudaxEnvironment): the environment to use for expansion
    - root_state (State): the initial state of the game at the root node
    - num_sims (int): the maximum number of nodes in the tree
    - max_depth (int): the maximum depth of the tree (used to size the rollout arrays). For safety, this can be set to num_sims
    - seed (int): random seed for initialization

    Returns:
    - params (MCTSParams): the initialized MCTS parameters
    - key (jax.random.PRNGKey): the updated random key
    '''
    transitions = -jnp.ones((num_sims, environment.num_actions), dtype=ACTION_DTYPE)
    states = jax.vmap(lambda i: root_state)(jnp.arange(num_sims))
    legal_actions = jnp.ones((num_sims, environment.num_actions), dtype=ACTION_DTYPE)
    visits = jnp.zeros((num_sims, environment.num_actions), dtype=ACTION_DTYPE)
    rewards = jnp.zeros((num_sims, environment.num_actions), dtype=REWARD_DTYPE)
    to_play = jnp.zeros((num_sims,), dtype=ACTION_DTYPE)

    legal_actions = legal_actions.at[0].set(root_state.legal_action_mask)
    to_play = to_play.at[0].set(root_state.current_player)
    
    params = MCTSParams(
        player_idx=root_state.current_player,
        node_num=0,
        transitions=transitions,
        states=states,
        legal_actions=legal_actions,
        visits=visits,
        rewards=rewards,
        to_play=to_play
    )
    
    return params, key

def next_state(mcts_params: MCTSParams, node_idx: int, action: int) -> int:
    '''
    Returns the node index of the child node resulting from taking the given action at the given node
    '''
    return mcts_params.transitions[node_idx, action]

def is_valid_node(mcts_params: MCTSParams, node_idx: int, action: int) -> bool:
    '''
    Returns whether the given action is valid at the given node (i.e. whether the child node has already been expanded)
    '''
    return mcts_params.transitions[node_idx, action] != -1

def select_action_ucb(mcts_params: MCTSParams, node_idx: int, key: jax.random.PRNGKey, c: float = 1.41) -> int:
    '''
    Selects an action at the given node using the UCB formula

    Params:
    - mcts_params (MCTSParams): the current MCTS parameters
    - node_idx (int): the index of the node to select an action from (the parent)
    - c (float): the exploration constant

    Returns:
    - action (int): the selected action
    '''
    visits = mcts_params.visits[node_idx]
    legal_actions = mcts_params.legal_actions[node_idx]

    # The rewards need to be flipped if the player to play at this node is not the root player
    rewards = jax.lax.select(
        mcts_params.to_play[node_idx] == mcts_params.player_idx,
        mcts_params.rewards[node_idx],
        -mcts_params.rewards[node_idx]
    )
    '''
    rewards = rewards if (player := root_player) else -rewards
    '''
    
    total_visits = jnp.sum(visits)
    
    # UCB formula
    ucb_values = jnp.where(
        visits > 0,
        rewards / visits + c * jnp.sqrt(jnp.log(total_visits) / visits),
        jnp.inf
    )

    ucb_values = jnp.where(legal_actions, ucb_values, -jnp.inf)

    logits = jnp.where(ucb_values == ucb_values.max(), 0.0, -jnp.inf)
    max_action = jax.random.categorical(key, logits=logits, axis=0).astype(ACTION_DTYPE)

    return max_action, ucb_values

def traverse_to_leaf(mcts_params: MCTSParams, max_depth: int, key: jax.random.PRNGKey) -> MCTSTraversal:
    '''
    Proceeds from the root node to a leaf node by selecting actions using UCB

    Params:
    - mcts_params (MCTSParams): the current MCTS parameters
    - max_depth (int): the maximum depth of the tree (used to size the rollout arrays)

    Returns:
    - rollout (MCTSRollout): the rollout information
    '''

    def body_fn(carry, i):
        node_idx, action, valid, key = carry

        key, subkey = jax.random.split(key)

        # If we've reached an unexpanded node, then all subsequent steps are invalid
        new_valid = valid & is_valid_node(mcts_params, node_idx, action)

        # Advance to the next node and select the next action. We don't need to worry about
        # the behavior here if the current node is invalid, since later all that information
        # will get discarded
        next_node_idx = next_state(mcts_params, node_idx, action)
        next_action, _ = select_action_ucb(mcts_params, next_node_idx, key)

        new_carry = (next_node_idx, next_action, new_valid, key)
        output = (node_idx, action, valid)

        return new_carry, output

    # Initialize the scan with the root node and an initial action
    node_idx, (action, _) = 0, select_action_ucb(mcts_params, 0, key)
    init_carry = (node_idx, action, True, key)
    (_, _, _, key), (nodes, actions, valid) = jax.lax.scan(body_fn, init_carry, jnp.arange(max_depth))

    rollout = MCTSTraversal(nodes=nodes, actions=actions, valid=valid)

    return rollout, key

def expand_leaf(mcts_params: MCTSParams, traversal: MCTSTraversal, environment: LudaxEnvironment, step_fn: callable, key: jax.random.PRNGKey) -> MCTSParams:
    '''
    Expands the leaf node reached by the given rollout

    Params:
    - mcts_params (MCTSParams): the current MCTS parameters
    - rollout (MCTSRollout): the rollout information
    - environment (LudaxEnvironment): the environment to use for expansion

    Returns:
    - mcts_params (MCTSParams): the updated MCTS parameters
    '''
    
    num_steps = len(traversal.nodes)
    last_valid_step = jnp.argmax(jnp.where(traversal.valid, jnp.arange(num_steps), 0))
    node_idx, action = traversal.nodes[last_valid_step], traversal.actions[last_valid_step]

    leaf_state = jax.tree_util.tree_map(lambda x: x[node_idx], mcts_params.states)
    next_state = step_fn(leaf_state, action.astype(ACTION_DTYPE))
    reward, key = evaluate_state(mcts_params, next_state, step_fn, key)

    # If the leaf was terminal, then we use the actual reward instead of a rollout reward
    # leaf_reward = (leaf_state.rewards[mcts_params.player_idx] + 1) / 2
    leaf_reward = leaf_state.rewards[mcts_params.player_idx] * REWARD_SCALE
    reward = jax.lax.select(leaf_state.terminated | leaf_state.truncated, leaf_reward, reward)

    expansion_node_idx = mcts_params.node_num + 1

    # The state needs to be updated using a tree map since it's a pytree (struct of arrays)
    updated_states = jax.tree_util.tree_map(lambda arr, new: arr.at[expansion_node_idx].set(new), mcts_params.states, next_state)

    num_sims = len(mcts_params.transitions)
    valid_node_idxs = jnp.where(traversal.valid, traversal.nodes, num_sims + 1)
    valid_actions = jnp.where(traversal.valid, traversal.actions, environment.num_actions + 1)

    # Increment the visit counts and rewards along the path (including the newly expanded node)
    updated_visits = mcts_params.visits.at[valid_node_idxs, valid_actions].add(1)
    updated_rewards = mcts_params.rewards.at[valid_node_idxs, valid_actions].add(reward)

    mcts_params = mcts_params._replace(
        visits=updated_visits,
        rewards=updated_rewards,
    )

    # Only update the transitions if the leaf node is not terminal
    mcts_params = jax.lax.cond(
        ~(leaf_state.terminated | leaf_state.truncated),
        lambda params: params._replace(
            node_num=expansion_node_idx,
            states=updated_states,
            legal_actions=params.legal_actions.at[expansion_node_idx].set(next_state.legal_action_mask),
            transitions=params.transitions.at[node_idx, action].set(expansion_node_idx),
            to_play=params.to_play.at[expansion_node_idx].set(next_state.current_player),

            # Update 9/11/2025 -- update the visit and reward for the expansion node as well
            visits=params.visits.at[expansion_node_idx].set(1),
            rewards=params.rewards.at[expansion_node_idx].set(reward),
        ),
        lambda params: params,
        mcts_params
    )

    return mcts_params, key

def evaluate_state(mcts_params: MCTSParams, state: State, step_fn: callable, key: jax.random.PRNGKey) -> float:
    '''
    Evaluates the given state by running a random rollout to a terminal state and returning the final reward
    '''

    def cond_fn(args):
        state, _ = args
        return ~(state.terminated | state.truncated).all()

    def body_fn(args):
        state, key = args
        key, subkey = jax.random.split(key)
        logits = jnp.log(state.legal_action_mask.astype(REWARD_DTYPE))
        action = jax.random.categorical(key, logits=logits, axis=0).astype(ACTION_DTYPE)
        state = step_fn(state, action)
        return state, key

    final_state, key = jax.lax.while_loop(cond_fn, body_fn, (state, key))

    # Return the reward for the player to play at the root, r \in [-1, 0, 1]
    reward = final_state.rewards[mcts_params.player_idx]
    reward *= REWARD_SCALE

    return reward, key


def uct_mcts_policy(environment, num_simulations=100, max_depth=20):
    step_fn = jax.jit(environment.step)

    def policy_single(state, key):
        params, key = initialize(environment, state, num_simulations, max_depth, key)

        def body_fn(i, carry):
            params, key = carry
            key, subkey = jax.random.split(key)
            rollout, key = traverse_to_leaf(params, max_depth, key)
            params, key = expand_leaf(params, rollout, environment, step_fn, key)

            return params, key

        params, key = jax.lax.fori_loop(0, num_simulations, body_fn, (params, key))
        return jnp.argmax(params.visits[0]).astype(ACTION_DTYPE)

    def policy_f(state_b, key):
        return jax.vmap(policy_single)(state_b, jax.random.split(key, state_b.rewards.shape[0]))

    return jax.jit(policy_f)



from importlib.resources import files

def _read_game(game_name: str) -> str:
    with files(__package__).joinpath(f"./{game_name}.ldx").open('r') as f:
        return f.read()

# Package a subset of default game implementations
connect_four = _read_game('connect_four')
connect_six = _read_game('connect_six')
dai_hasami_shogi = _read_game('dai_hasami_shogi')
english_draughts = _read_game('english_draughts')
english_draughts_hex = _read_game('english_draughts_hex')
gridworld = _read_game('gridworld')
hasami_shogi = _read_game('hasami_shogi')
hex = _read_game('hex')
hop_through = _read_game('hop_through')
gomoku = _read_game('gomoku')
pente = _read_game('pente')
reversi = _read_game('reversi')
test = _read_game('test')
tic_tac_toe = _read_game('tic_tac_toe')
yavalath = _read_game('yavalath')
yavalax = _read_game('yavalax')
wolf_and_sheep = _read_game('wolf_and_sheep')

# List of all games available in the package
__all__ = [
    "connect_four",
    "connect_six",
    "dai_hasami_shogi",
    "english_draughts",
    "english_draughts_hex",
    "gridworld",
    "hasami_shogi",
    "hex",
    "hop_through",
    "gomoku",
    "pente",
    "reversi",
    "test",
    "tic_tac_toe",
    "yavalath",
    "yavalax",
    "wolf_and_sheep",
]



import jax
import jax.numpy as jnp
from . import BIG, SMALL, zero_heuristic
from ludax.config import ACTION_DTYPE, REWARD_DTYPE


def random_policy():
    def random_policy_f(state_b, key):
        """
        Randomly select an action from the legal actions.
        :param state_b: Current state of the game.
        :param step_b: Function to step the game forward.
        :param key: JAX PRNG key for random number generation.
        :return: Selected action.
        """
        legal_action_mask = state_b.legal_action_mask
        logits = jnp.log(legal_action_mask.astype(REWARD_DTYPE))
        return jax.random.categorical(key, logits=logits, axis=1).astype(ACTION_DTYPE)

    return jax.jit(random_policy_f)


def one_ply_policy(step_b, heuristic=zero_heuristic):
    def one_step_lookahead_f(state_b, key):
        """
        Pick the move with the best one‑ply value for the *current* mover (i.e. maximise next_state.mover_reward).
        All else equal, consider heuristic.
        """
        batch_size, num_actions = state_b.legal_action_mask.shape

        # (batch, num_actions) repeat each action for each state in the batch
        all_actions = jnp.broadcast_to(jnp.arange(num_actions), (batch_size, num_actions))
        actions_flat = all_actions.reshape(-1)

        # Repeat every state `num_actions` times
        state_flat = jax.tree_util.tree_map(lambda x: jnp.repeat(x, num_actions, axis=0), state_b)

        # Step every (state, action) pair in one call
        flat_next_state = step_b(state_flat, actions_flat.astype(ACTION_DTYPE))

        # Sum the next_state.mover_reward with the heuristic
        # Note: even in games where the players always alternate, this is necessary since the terminal state doesn't
        # switch the current player, and thus the heuristic will be with respect to wrong player.
        # In fact `polarity = jnp.where(next_state.terminated, 1, -1)` is equivalent in Hex
        polarity = jnp.where(flat_next_state.game_state.current_player == state_flat.game_state.current_player, 1.0, -1.0)
        # debug_polarity = jnp.where(flat_next_state.terminated, 0.0, polarity)
        # jax.debug.print("polarity: {polarity}", polarity=debug_polarity, ordered=True)

        # Add some noise to the heuristic to break ties non-deterministically
        noise = jax.random.uniform(key, shape=flat_next_state.mover_reward.shape, minval=-SMALL, maxval=SMALL)

        action_values_flat = BIG * flat_next_state.mover_reward + polarity * heuristic(flat_next_state) + noise

        # Mask out illegal actions. Som
        action_values = action_values_flat.reshape(batch_size, num_actions)
        action_values = jnp.where(state_b.legal_action_mask, action_values, -jnp.inf)

        return jnp.argmax(action_values, axis=1).astype(ACTION_DTYPE)

    return jax.jit(one_step_lookahead_f)


from typing import Optional

import jax
import jax.numpy as jnp
from lark import Lark

import ludax
from .config import Array, PRNGKey, State, EMPTY, TRUE, MAX_STEP_COUNT, BOARD_DTYPE, ACTION_DTYPE, REWARD_DTYPE
from .game_info import GameInfoExtractor
from .game_parser import GameRuleParser


class LudaxEnvironment():
    def __init__(self,
                 game_path: str = None,
                 game_str: str = None):
        super().__init__()

        assert game_path is not None or game_str is not None, "Must provide either a game path or a game string!"
        assert game_path is None or game_str is None, "Cannot provide both a game path and a game string!"
        
        if game_path is not None:
            with open(game_path, 'r') as f:
                game_str = f.read()

        parser = Lark(ludax.grammar, start='game')

        game_tree = parser.parse(game_str)
        self.game_info, self.rendering_info = GameInfoExtractor()(game_tree)
        game_rules = GameRuleParser(self.game_info).transform(game_tree)

        self.game_state_cls = self.game_info.game_state_class

        self.obs_shape = self.game_info.observation_shape
        self.board_size = self.game_info.board_size
        self.num_piece_types = self.game_info.num_piece_types

        self.num_actions = game_rules['action_size']
        self._initialize_board = game_rules['start_rules']
        self._apply_action = game_rules['apply_action_fn']
        self._update_info = game_rules['addl_info_fn']
        self._get_legal_action_mask = game_rules['legal_action_mask_fn']
        self._apply_effects = game_rules['apply_effects_fn']
        self._get_phase_idx = game_rules['next_phase_fn']
        self._get_next_player = game_rules['next_player_fn']

        self._get_winner = game_rules['end_rules']
        self._get_terminal = lambda board, player: (board != EMPTY).all()

    def init(self, rng: PRNGKey) -> State:

        # Temporarily hard-coding the init of the game state
        temp_current_player = BOARD_DTYPE(0)
        game_state = self.game_state_cls(
            board=jnp.ones((self.num_piece_types, self.board_size), dtype=BOARD_DTYPE) * EMPTY,
            legal_action_mask=jnp.ones((self.num_actions,), dtype=jnp.bool_),
            current_player=temp_current_player,
            phase_idx=BOARD_DTYPE(0),
            phase_step_count=BOARD_DTYPE(0),
            previous_actions=jnp.array([-1, -1, -1], dtype=ACTION_DTYPE),
        )

        # Initialize the board using the game rules
        game_state = self._initialize_board(game_state)

        # Compute the actual starting player
        current_player = self._get_next_player(game_state)
        game_state = game_state._replace(current_player=current_player)

        game_state = self._update_info(game_state, -1)

        legal_action_mask = self._get_legal_action_mask(game_state).astype(jnp.bool_)
        game_state = game_state._replace(legal_action_mask=legal_action_mask)

        state = State(
            game_state=game_state,
            legal_action_mask=legal_action_mask,
            current_player=current_player
        )

        return state

    def step(self, state: State, action: Array, key: Optional[Array] = None) -> State:
        """
        The basic environment step function, largely taken from the PGX implementation
        (https://github.com/sotetsuk/pgx/blob/main/pgx/core.py) and used under the
        Apache-2.0 license.
        """
        is_illegal = ~state.legal_action_mask[action]
        current_player = state.current_player

        # If the state is already terminated or truncated, environment does not take usual step,
        # but return the same state with zero-rewards for all players
        state = jax.lax.cond(
            (state.terminated | state.truncated),
            lambda: state.replace(rewards=jnp.zeros_like(state.rewards)),  # type: ignore
            lambda: self._step(state, action, key),  # type: ignore
        )

        # Taking illegal action leads to immediate game terminal with negative reward
        state = jax.lax.cond(
            is_illegal,
            lambda: self._step_with_illegal_action(state, current_player),
            lambda: state,
        )

        # All legal_action_mask elements are **TRUE** at terminal state
        # This is to avoid zero-division error when normalizing action probability
        # Taking any action at terminal state does not give any effect to the state
        state = jax.lax.cond(
            state.terminated,
            lambda: state.replace(legal_action_mask=jnp.ones_like(state.legal_action_mask)),  # type: ignore
            lambda: state,
        )

        # NOTE: unlike PGX, we don't compute the observation inside the step function, instead leaving it
        # to, for example, the RL training loop using env._observe(state, player_id)

        return state

    def _step(self, state: State, action: Array, key) -> State:
        # Currently, Ludax games don't have any stochastic elements, so we don't use the key
        del key

        # The game state is separate from the "environment state" and contains
        # only the information necessary to progress the game
        game_state = state.game_state

        # Track the player who started the turn
        original_player = game_state.current_player

        # Compute the new board state
        game_state = self._apply_action(game_state, action)

        # Compute any 'additional information' required by later steps. For instance, connected
        # components
        game_state = self._update_info(game_state, action)

        # Compute the new phase index
        new_phase_idx, phase_step_count = self._get_phase_idx(game_state)
        game_state = game_state._replace(phase_idx=new_phase_idx, phase_step_count=phase_step_count)

        # Use the new phase and the global player offset to determine the next player
        next_player = self._get_next_player(game_state)
        game_state = game_state._replace(current_player=next_player)

        # Apply any effects that occur at the end of the turn. The effects functions take in the
        # original player so they can correctly handle "extra turns" given to the mover / opponent
        game_state = self._apply_effects(game_state, original_player)

        # Compute the legal action mask for the upcoming player (which is used in some end conditions)
        # TODO: should this be stored separately for each player?
        new_legal_action_mask = self._get_legal_action_mask(game_state).astype(jnp.bool_)
        game_state = game_state._replace(legal_action_mask=new_legal_action_mask)
        state = state.replace(legal_action_mask=new_legal_action_mask)

        # Use the new board to compute the winner, terminal, and rewards -- but consider the
        # current player to be the "original" player so they get credit for winning on their turn
        winners, terminated = self._get_winner(game_state._replace(current_player=original_player))
        terminal_rewards = jnp.where((winners == EMPTY).all(), jnp.zeros_like(winners), jnp.where(winners, 1, -1)).astype(REWARD_DTYPE)
        rewards = jax.lax.select(terminated, terminal_rewards, jnp.zeros_like(terminal_rewards))

        mover_reward = rewards[original_player]
        state = state.replace(winners=winners, rewards=rewards, terminated=terminated, mover_reward=mover_reward)

        # If the terminated, then reset the current player back to the original player so that
        # "mover reward" refers to the correct player
        game_state = jax.lax.cond(
            terminated,
            lambda: game_state._replace(current_player=original_player),
            lambda: game_state
        )

        # Update the game state in the state
        state = state.replace(game_state=game_state, current_player=game_state.current_player)

        # Increment the global step count
        state = state.replace(global_step_count=state.global_step_count + 1, truncated=state.global_step_count >= MAX_STEP_COUNT)

        return state
    
    def observe(self, state: State, player_id: jax.Array) -> jax.Array:
        """
        Convert a multi-piece board into a boolean (..., width, height, 2 * num_pieces) tensor.
        board[..., p, i] == -1  → empty square
        board[..., p, i] == 0   → square occupied by white's piece of type `p`
        board[..., p, i] == 1   → square occupied by black's piece of type `p`
        
        The resulting channels are grouped: all of the current player's pieces, followed by all of the opponent's pieces.
        """
        board = state.game_state.board
        
        # Assuming observation_shape is updated to (width, height, 2 * num_pieces)
        width, height = self.game_info.observation_shape[0], self.game_info.observation_shape[1]
        
        # Extract num_pieces from the board's shape: (..., num_pieces, width * height)
        num_pieces = board.shape[-2] 

        # 1. Reshape the flat spatial dimension to 2D
        # Resulting shape: (..., num_pieces, width, height)
        board2d = board.reshape(*board.shape[:-1], width, height)

        # 2. Expand player_id for broadcasting against the 4 trailing dimensions
        # Shape becomes (..., 1, 1, 1) to broadcast over (num_pieces, width, height)
        p_id = player_id[..., None, None, None]

        # 3. Create boolean masks
        # Both will have shape: (..., num_pieces, width, height)
        # Empty squares (-1) resolve to False automatically since p_id is always 0 or 1
        my_pieces = (board2d == p_id)
        op_pieces = (board2d == jnp.abs(1 - p_id))

        # 4. Move the `num_pieces` axis to the end to act as channels
        # Using -3 ensures this works seamlessly whether the input is batched or unbatched
        # Resulting shape: (..., width, height, num_pieces)
        my_pieces_channels = jnp.moveaxis(my_pieces, -3, -1)
        op_pieces_channels = jnp.moveaxis(op_pieces, -3, -1)

        # 5. Concatenate along the channel dimension
        # Resulting shape: (..., width, height, 2 * num_pieces)
        obs = jnp.concatenate((my_pieces_channels, op_pieces_channels), axis=-1)

        return jax.lax.stop_gradient(obs)  # Taken from PGX

    @property
    def _illegal_action_penalty(self) -> float:
        """
        Negative reward given when illegal action is selected
        """
        return -1.0
    
    def _step_with_illegal_action(self, state: State, loser: Array) -> State:
        penalty = self._illegal_action_penalty
        reward = jnp.ones_like(state.rewards) * (-1 * penalty) * (self.game_info.num_players - 1)
        reward = reward.at[loser].set(penalty)
        return state.replace(rewards=reward, terminated=TRUE)  # type: ignore



import math
import re
import svgwrite
from svgwrite import cm, mm

import jax.numpy as jnp
import numpy as np

from ..config import Shapes, PieceShapes, P1, P2, EMPTY, RENDER_CONFIG, BOARD_DTYPE


class InteractiveBoardHandler():
    '''
    Facilitates an interactive version of a game environment by generating an SVG board that can be clicked on
    and mapping clicks to actions in the game environment
    '''
    def __init__(self,
                 game_info,
                 rendering_info,
                 render_config=RENDER_CONFIG):
        
        self.game_info = game_info
        self.rendering_info = rendering_info
        self.render_config = render_config
        self.cell_size = render_config['cell_size']
        self.padding = self.cell_size / 2
        self.rendered_svg = ""
        self.animation_snippet = '<animate attributeName="opacity" values="1;0.33;1" dur="1s" calcMode="paced" fill="freeze" />'
        

        if self.game_info.board_shape == Shapes.SQUARE or self.game_info.board_shape == Shapes.RECTANGLE:
            self.cell_size *= 2
            height, width = self.game_info.board_dims
            self.action_indices = [(y, x) for y in range(height) for x in range(width)]
            self._action_map = {idx: i for i, idx in enumerate(self.action_indices)}
            self.action_to_pixel = lambda action: self._grid_to_pixel(self.action_indices[action])
            self.pixel_to_action = lambda pixel: self._action_map[self._pixel_to_grid(pixel)]
            self.get_cell_vertices = lambda position: self._get_grid_vertices(position)

            # Compute the total width and height of the board
            self.total_width = width * self.cell_size
            self.total_height = height * self.cell_size
        
        elif self.game_info.board_shape == Shapes.HEXAGON:
            self.orientation = render_config['hexagon_orientation']

            # Compute the ordered sequence or (q, r, s) indices that correspond to horizontal scans over the board
            diameter = self.game_info.hex_diameter
            n = diameter // 2
            self.action_indices = [(q, r, s) for r in range(-n, n+1) for s in range(n, -(n+1), -1) for q in range(-n, n+1) if q + r + s == 0]
            self._action_map = {idx: i for i, idx in enumerate(self.action_indices)}
            self.action_to_pixel = lambda action: self._hex_to_pixel(self.action_indices[action])
            self.pixel_to_action = lambda pixel: self._action_map[self._pixel_to_hex(pixel)]
            self.get_cell_vertices = lambda position: self._get_hexagon_vertices(position)

            # Compute the total width and height of the board
            if self.orientation == 'flat':
                hex_width = 2 * self.cell_size
                hex_height = math.sqrt(3) * self.cell_size

                half_cells_across = diameter // 2
                full_cells_across = diameter - half_cells_across
                
                self.total_width = (hex_width * full_cells_across) + (hex_width * half_cells_across / 2)
                self.total_height = hex_height * diameter

            elif self.orientation == 'pointy':
                hex_width = math.sqrt(3) * self.cell_size
                hex_height = 2 * self.cell_size

                half_cells_across = diameter // 2
                full_cells_across = diameter - half_cells_across

                self.total_width = hex_width * diameter
                self.total_height = (hex_height * full_cells_across) + (hex_height * half_cells_across / 2)

        elif self.game_info.board_shape == Shapes.HEX_RECTANGLE:
            # The canonical 'hex rectangle' is always 'pointy'
            self.orientation = 'pointy'
            height, width = self.game_info.board_dims
            self.action_indices = [(y, x) for y in range(height) for x in range(width)]
            self._action_map = {idx: i for i, idx in enumerate(self.action_indices)}
            self.action_to_pixel = lambda action: self._hex_rectangle_to_pixel(self.action_indices[action])
            self.pixel_to_action = lambda pixel: self._action_map[self._pixel_to_hex_rectangle(pixel)]
            self.get_cell_vertices = lambda position: self._get_hexagon_vertices(position)


            self.hex_width = math.sqrt(3) * self.cell_size
            self.hex_height = 2 * self.cell_size

            # Each row adds a half-cell to the width
            self.total_width = (self.hex_width * width) + ((height - 1) * self.hex_width / 2)

            half_cells_up = height // 2
            full_cells_up = height - half_cells_up
            self.total_height = (self.hex_height * full_cells_up) + (self.hex_height * half_cells_up / 2)

        else:
            raise ValueError(f"Unknown board shape: {self.game_info.board_shape}")
        
        # Add padding to the total width and height
        self.total_width += self.padding
        self.total_height += self.padding

        self.draw_fns = {
            PieceShapes.CIRCLE: self._draw_circle,
            PieceShapes.SQUARE: self._draw_square,
            PieceShapes.TRIANGLE: self._draw_triangle,
            PieceShapes.STAR: self._draw_star,
            PieceShapes.DIAMOND: self._draw_diamond,
        }

        # Compute region masks using a dummy initial state
        self.region_masks = {}
        self._compute_region_masks()

    def _compute_region_masks(self):
        """
        Compute region masks by running each region_mask_fn on a dummy initial state.
        Stores the resulting masks indexed by region name.
        """
        if self.game_info.region_mask_fns is None or self.game_info.num_regions == 0:
            return
        
        # Create a dummy state with an empty board to evaluate region mask functions
        # Most region functions just need the board shape, not actual piece positions
        from collections import namedtuple
        
        # Create a minimal dummy state with required attributes
        DummyState = namedtuple('DummyState', ['board', 'current_player'])
        dummy_board = jnp.ones((self.game_info.num_piece_types, self.game_info.board_size), dtype=BOARD_DTYPE) * EMPTY
        dummy_state = DummyState(board=dummy_board, current_player=BOARD_DTYPE(0))
        
        # Evaluate each region mask function and store the result
        region_colors = self.render_config.get('region_colors', [])
        for idx, (region_name, mask_fn) in enumerate(zip(self.game_info.region_names, self.game_info.region_mask_fns)):
            try:
                mask = mask_fn(dummy_state)
                # Convert to numpy array for easier indexing during rendering
                mask_array = np.array(mask).astype(bool)
                color = region_colors[idx % len(region_colors)] if region_colors else self.render_config['light_blue']
                self.region_masks[region_name] = {
                    'mask': mask_array,
                    'color': color
                }
            except Exception as e:
                # If evaluation fails (e.g., mask depends on game state), skip this region
                print(f"Warning: Could not compute region mask for '{region_name}': {e}")
                continue

    def _get_cell_fill_color(self, cell_index):
        """
        Determine the fill color for a cell based on which regions it belongs to.
        Returns the color of the first matching region, or the default light_blue.
        """
        for region_name, region_info in self.region_masks.items():
            if cell_index < len(region_info['mask']) and region_info['mask'][cell_index]:
                return region_info['color']
        return self.render_config["light_blue"]

    def render_legend(self):
        """
        Generate an HTML legend showing region names and their colors.
        Returns an HTML string that can be inserted into the page.
        """
        if not self.region_masks:
            return ""
        
        legend_items = []
        for region_name, region_info in self.region_masks.items():
            color = region_info['color']
            formatted_name = region_name.replace('"', '').replace('_', ' ').title()

            # Create a colored box and region name
            item = f'''
            <div style="display: flex; align-items: center; margin: 5px 0;">
                <div style="width: 20px; height: 20px; background-color: {color}; 
                            border: 1px solid #666; margin-right: 10px;"></div>
                <span>{formatted_name}</span>
            </div>
            '''
            legend_items.append(item)
        
        legend_html = f'''
        <div style="margin-top: 15px; padding: 10px; border: 1px solid #ccc; 
                    border-radius: 5px; background-color: #f9f9f9;">
            <h3 style="margin: 0 0 10px 0; font-size: 20px;">Regions</h3>
            {''.join(legend_items)}
        </div>
        '''
        return legend_html

    def _grid_to_pixel(self, grid_point):
        '''
        Converts grid coordinates to pixel coordinates
        '''

        # Grid points take the form (y, x)
        x, y = grid_point[1] * self.cell_size, grid_point[0] * self.cell_size

        # Offset by just the padding amount and center in the cell
        x += self.padding / 2 + self.cell_size / 2
        y += self.padding / 2 + self.cell_size / 2
        
        return x, y
    
    def _pixel_to_grid(self, pixel_point):
        '''
        Converts pixel coordinates to grid coordinates
        '''

        # Pixel points take the form (x, y)
        x, y = pixel_point

        # Reverse the padding offset
        x -= self.padding / 2
        y -= self.padding / 2

        # Convert to grid coordinates
        x = x // self.cell_size
        y = y // self.cell_size

        return y, x

    def _hex_to_pixel(self, hex_point):
        '''
        Converts hexagonal coordinates to pixel coordinates
        '''

        q, r, s = hex_point

        if self.orientation == 'flat':
            x = self.cell_size * (3/2) * q
            y = self.cell_size * (math.sqrt(3)/2 * q + math.sqrt(3) * r)

        elif self.orientation == 'pointy':
            x = self.cell_size * (math.sqrt(3) * q + math.sqrt(3)/2 * r)
            y = self.cell_size * (3/2) * r

        # Offset the coordinates so that the center of the board is at (0, 0)
        x += self.total_width / 2
        y += self.total_height / 2

        return x, y
    
    def _pixel_to_hex(self, pixel_point):
        x, y = pixel_point

        # Reverse the offset applied in _hex_to_pixel
        x -= self.total_width / 2
        y -= self.total_height / 2

        # Apply the inverse of the hex_to_pixel conversion
        q = ((x * math.sqrt(3) / 3) - (y / 3)) / 50
        r = (y * 2 / 3) / 50
        s = -q - r

        # Round the cube coordinates to the nearest hexagon
        q_round, r_round, s_round = round(q), round(r), round(s)
        q_diff, r_diff, s_diff = abs(q_round - q), abs(r_round - r), abs(s_round - s)

        if q_diff > r_diff and q_diff > s_diff:
            q_round = -r_round - s_round

        elif r_diff > s_diff:
            r_round = -q_round - s_round

        else:
            s_round = -q_round - r_round

        return q_round, r_round, s_round
    
    def _hex_rectangle_to_pixel(self, hex_point):
        x, y = hex_point[1] * self.hex_width, hex_point[0] * self.hex_height * 0.75

        x += hex_point[0] * (self.hex_width / 2)

        x += (self.padding / 2) + (self.hex_width / 2)
        y += (self.padding / 2) + (self.hex_height / 2)

        return x, y
    
    def _pixel_to_hex_rectangle(self, pixel_point):
        x, y = pixel_point

        x -= (self.padding / 2) + (self.hex_width / 2)
        y -= (self.padding / 2) + (self.hex_height / 2)

        q = ((x * math.sqrt(3) / 3) - (y / 3)) / self.cell_size
        r = (y * 2 / 3) / self.cell_size
        q_round, r_round = round(q), round(r)

        return r_round, q_round

    def _get_grid_vertices(self, position):
        '''
        Returns the vertices of a square at the given position
        '''
        x, y = position

        points = [
            (x - self.cell_size / 2, y - self.cell_size / 2),
            (x + self.cell_size / 2, y - self.cell_size / 2),
            (x + self.cell_size / 2, y + self.cell_size / 2),
            (x - self.cell_size / 2, y + self.cell_size / 2)
        ]

        return points

    def _get_hexagon_vertices(self, position):
        '''
        Returns the vertices of a hexagon at the given position
        '''
        x, y = position

        if self.orientation == 'flat':
            width = 2 * self.cell_size
            height = math.sqrt(3) * self.cell_size

            points = [
                (x - 0.25 * width, y - 0.5 * height),
                (x + 0.25 * width, y - 0.5 * height),
                (x + 0.5 * width, y),
                (x + 0.25 * width, y + 0.5 * height),
                (x - 0.25 * width, y + 0.5 * height),
                (x - 0.5 * width, y)
            ]

        elif self.orientation == 'pointy':
            width = math.sqrt(3) * self.cell_size
            height = 2 * self.cell_size

            points = [
                (x - 0.5 * width, y - 0.25 * height),
                (x, y - 0.5 * height),
                (x + 0.5 * width, y - 0.25 * height),
                (x + 0.5 * width, y + 0.25 * height),
                (x, y + 0.5 * height),
                (x - 0.5 * width, y + 0.25 * height)
            ]

        return points

    def _draw_circle(self, drawing, center, size, fill, stroke, stroke_width, class_=""):
        drawing.add(drawing.circle(center=center, r=size, fill=fill, stroke=stroke, stroke_width=stroke_width, class_=class_))

    def _draw_square(self, drawing, center, size, fill, stroke, stroke_width, class_=""):
        top_left = (center[0] - size, center[1] - size)
        drawing.add(drawing.rect(insert=top_left, size=(2 * size, 2 * size), fill=fill, stroke=stroke, stroke_width=stroke_width, class_=class_))

    def _draw_triangle(self, drawing, center, size, fill, stroke, stroke_width, class_=""):
        points = []
        angle_step = 2 * math.pi / 3

        for i in range(3):
            angle = i * angle_step - math.pi / 2 
            x = center[0] + size * math.cos(angle)
            y = center[1] + size * math.sin(angle)
            points.append((x, y))

        drawing.add(drawing.polygon(points, fill=fill, stroke=stroke, stroke_width=stroke_width, class_=class_))

    def _draw_star(self, drawing, center, size, fill, stroke, stroke_width, num_points=5, class_=""):
        points = []
        angle_step = 2 * math.pi / (2 * num_points)

        for i in range(2 * num_points):
            radius = size if i % 2 == 0 else (size / 2)
            angle = i * angle_step - math.pi / 2 
            x = center[0] + radius * math.cos(angle)
            y = center[1] + radius * math.sin(angle)
            points.append((x, y))

        drawing.add(drawing.polygon(points, fill=fill, stroke=stroke, stroke_width=stroke_width, class_=class_))

    def _draw_diamond(self, drawing, center, size, fill, stroke, stroke_width, class_=""):
        points = [
            (center[0], center[1] - size),
            (center[0] + size, center[1]),
            (center[0], center[1] + size),
            (center[0] - size, center[1])
        ]
        drawing.add(drawing.polygon(points, fill=fill, stroke=stroke, stroke_width=stroke_width, class_=class_))


    def render(self, state, add_button=True, show_legal_actions=True, legal_actions=None, selected_action=None):
        board = state.game_state.board
        if legal_actions is None and show_legal_actions:
            legal_actions = state.legal_action_mask

        # The third index of previous_actions stores the last action taken (regardless of player)
        last_action = state.game_state.previous_actions[-1]

        self.rendered_svg = self.render_fn(board, legal_actions=legal_actions, add_button=add_button, last_action=last_action, selected_action=selected_action)


    def render_fn(self, board, legal_actions=None, add_button=True, last_action=None, selected_action=None):
        # Initialize drawing and draw boarder
        drawing = svgwrite.Drawing(size=(self.total_width, self.total_height), id="game_board")

        drawing.add(drawing.rect(insert=(0, 0), size=(self.total_width, self.total_height), stroke='black', stroke_width=1, fill='none'))

        # Iterate over the different kinds of pieces
        for piece_id, sub_board in enumerate(board):
            piece_name = self.game_info.piece_names[piece_id]

            for i, occupant in enumerate(sub_board):
                position = self.action_to_pixel(i)
                vertices = self.get_cell_vertices(position)

                # Draw the cell only in the first sub-board (to avoid overdrawing)
                if piece_id == 0:
                    cell_fill = self._get_cell_fill_color(i)
                    drawing.add(drawing.polygon(vertices, fill=cell_fill, stroke=self.render_config["light_grey"], stroke_width=1))

                if last_action is not None and i == last_action:
                    cls = "last-action"
                else:
                    cls = "other-action"

                # Draw the piece (if present)
                draw_fn = self.draw_fns[self.rendering_info.piece_shape_mapping[piece_name]]
                if occupant == P1:
                    fill_color = self.render_config[self.rendering_info.color_mapping['P1']]
                    draw_fn(drawing, center=position, size=self.render_config['piece_radius'], fill=fill_color, stroke=self.render_config['dark_grey'], stroke_width=1, class_=cls)
                elif occupant == P2:
                    fill_color = self.render_config[self.rendering_info.color_mapping['P2']]
                    draw_fn(drawing, center=position, size=self.render_config['piece_radius'], fill=fill_color, stroke=self.render_config['dark_grey'], stroke_width=1, class_=cls)



                # Draw the legal action mask
                if legal_actions is not None and legal_actions[i]:
                    drawing.add(drawing.circle(center=position, r=self.render_config['legal_radius'], fill=self.render_config['purple'], stroke=self.render_config['dark_grey'], stroke_width=1))

        # Draw a selection ring around the currently selected piece
        if selected_action is not None:
            position = self.action_to_pixel(selected_action)
            ring_radius = self.render_config['piece_radius'] + 8
            drawing.add(drawing.circle(center=position, r=ring_radius, fill='none', stroke='#f0a500', stroke_width=3))

        # Add an invisible rectangle to capture the user's clicks
        if add_button:
            drawing.add(drawing.rect(insert=(0, 0), size=(self.total_width, self.total_height), class_="btn", onclick="handleClick(event)"))

        drawstr = drawing.tostring()
        

        # Insert animation into shapes for P1 pieces
        shape_pattern = re.compile(r'(<(circle|rect|polygon)[^>]*>)')
        def add_animation(match):
            shape_tag = match.group(1)
            shape_id = shape_tag.split(" ")[0][1:]

            # Insert animation before closing tag
            if "last-action" in shape_tag:
                return shape_tag.replace('/>', f'> {self.animation_snippet} </{shape_id}>')
            else:
                return shape_tag

        drawstr = shape_pattern.sub(add_animation, drawstr)


        return drawstr